/*
* Copyright (c) 2003, University of Maryland 
* Copyright (c) 2003, Center for Satellite and Hybrid Communication Networks 
* All rights reserved.
*
* Redistribution and use in source and binary forms, with or without 
* modification, are permitted provided that the following conditions 
* are met:
*
*   * Redistributions of source code must retain the above copyright notice, 
*     this list of conditions,the authors and the following disclaimer.
*   * Redistributions in binary form must reproduce the above copyright notice,
*     this list of conditions,the authors and the following disclaimer in
*     the documentation and/or other materials provided with the distribution.
*   * Neither the name of the University of Maryland/Center for Satellite
*     and Hybrid Communication Networks nor the names of its 
*     contributors may be used to endorse or promote products derived from 
*     this software without specific prior written permission.
*
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
* AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
* IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
* ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
* LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
* CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
* SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
* INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
* ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
* POSSIBILITY OF SUCH DAMAGE.
*
*/

/* Authors: 
*     Jonathan Polley (jpolley@isr.umd.edu),
*     Dionysus Blazakis (dblaze@isr.umd.edu),
*     Jonathan McGee (mcgee@isr.umd.edu),
*     Daniel Rusk (ruskd@isr.umd.edu),
*     Karl Seamon (kks@isr.umd.edu),
*     Manish Karir (karir@isr.umd.edu)
*/

#ifndef ATEMU_ATMEGA128_H
#define ATEMU_ATMEGA128_H

#include "defs.h"
#include "device.h"
#include "table.h"
#include "stack.h"

enum {
    ATEMU_MODE_DECODE,
    ATEMU_MODE_EXECUTE
};

typedef struct atmega_chip {
    int id;

    uint8_t *sram;
    uint32_t sram_bytes;

    uint32_t io_offset; /* offset to io space for in/out and friends */
    uint32_t io_last;   /* end of io space */

    uint16_t *flash;
    uint32_t flash_words;

    uint8_t pc_bits;    /* number of bits PC is */
    uint32_t pc_mask;   /* AND mask to bound flash access */

    /* docs say pc is 16 or 22 bits */
    uint32_t pc;
    uint32_t oldpc;
    uint32_t start_vector;

    uint8_t *sreg;                      /* Status Register */
    uint8_t sreg_io;                    /* SREG IO Register address */

    #define ATEMU_SREG_I       0x80    /* Global Interrupt Enable */
    #define ATEMU_SREG_T       0x40    /* Bit Copy Storage */
    #define ATEMU_SREG_H       0x20    /* Half Carry Flag */
    #define ATEMU_SREG_S       0x10    /* Sign Bit, S = N ^ V */
    #define ATEMU_SREG_V       0x08    /* 2's Complement Overflow Flag */
    #define ATEMU_SREG_N       0x04    /* Negative Flag */
    #define ATEMU_SREG_Z       0x02    /* Zero Flag */
    #define ATEMU_SREG_C       0x01    /* Carry Flag */

    uint8_t *register_file;            /* R0 - R31 */

    /* X, Y, and Z double-byte registers */
    #define ATEMU_REG_X       0x1a
    #define ATEMU_REG_X_LOW   0x1a
    #define ATEMU_REG_X_HIGH  0x1b
    #define ATEMU_REG_Y       0x1c
    #define ATEMU_REG_Y_LOW   0x1c
    #define ATEMU_REG_Y_HIGH  0x1d
    #define ATEMU_REG_Z       0x1e
    #define ATEMU_REG_Z_LOW   0x1e
    #define ATEMU_REG_Z_HIGH  0x1f

    uint8_t *spl;
    uint8_t *sph;

    uint8_t spl_io;                     /* IO Space address of Stack pointer */
    uint8_t sph_io;                     /* IO Space address of Stack pointer */

    uint8_t *rampz;                     /* RAM Page */
    uint8_t rampz_io;                   /* IO Space address of rampz */

    uint8_t eind_io;                    /* IO Address of EIND */
    uint8_t *eind;

    uint64_t cycles;
    uint64_t oldcycles;

    /* internal registers */
    uint16_t instr;
    uint8_t th_nibble, tl_nibble, bh_nibble, bl_nibble, t_byte, b_byte;

    uint8_t reg_index_a;
    uint8_t reg_index_b;
    uint8_t a;
    uint8_t b;
    uint32_t addr;

    /* Bitmask of all interrupts (1 in bit position if active) */
    uint64_t interrupts;

    /* TRUE if the processor is currently executing a SLEEP op-code */
    int am_sleeping:1;

    /* io manager state */
    register_obj *register_head[256];
    interrupt_obj *interrupt_head;
    device_obj *device_head;
    clock_obj *clock_head;

    int mode;

    call_stack_t *call_stack;
    int stack_changed;

    table_t *breaktable;
    table_t *watchtable;
    table_t *notifytable;
    int watchhit;
} atmega_chip;

typedef struct{
    uint32_t addr;
}break_t;

typedef struct{
    uint32_t addr;
    int id;
    uint8_t value;
}watch_t;

int breakcomp(const void *a, const void *b);
int watchcomp(const void *a, const void *b);

void atmega_branch_skip_instruction( atmega_chip *chip );

error_t atmega_read_sram( atmega_chip *chip, uint16_t addr,
    uint8_t *value );
error_t atmega_write_sram( atmega_chip *chip, uint16_t addr,
    uint8_t value );

int atemu_load_file( atmega_chip ***chips, char *fname, char *config );
error_t atmega_load_chip(atmega_chip **chips, int i, char *config, char *srec);
error_t atmega_create_chip( atmega_chip **ret_chip, char *conf );
error_t atmega_destroy_chip( atmega_chip **chip );
error_t atmega_load_flash( atmega_chip *chip, char *filename );
error_t atmega_step( atmega_chip *chip );

/* Cycle count to halt execution at */
extern uint64_t halttime;

#endif /* ATEMU_ATMEGA128_H */
